From lrust.lifetime Require Export primitive.
From lrust.lifetime Require Import faking.
From iris.algebra Require Import csum auth frac gmap agree gset numbers excl.
From iris.proofmode Require Import tactics.
From iris.prelude Require Import options.

Section creation.
Context `{!invGS Σ, !lftG Σ Lat userE, LatBottom Lat}.
Implicit Types κ : lft.

Local Set Default Proof Using "Type*".

Lemma lft_kill (I : gmap lft lft_names) (K K' : gset lft) (κ : lft) (Vs : lft → Lat) :
  (∀ κ', κ' ∈ dom I → κ' ⊆ κ → Vs κ ⊑ Vs κ') →
  let Iinv := (
    own_ilft_auth I ∗
    ([∗ set] κ' ∈ K, lft_inv_dead κ' (Vs κ')) ∗
    ([∗ set] κ' ∈ K', lft_inv_alive κ' (Vs κ')))%I in
  (∀ κ', κ' ∈ dom I → κ ⊂ κ' → κ' ∈ K) →
  (∀ κ', κ' ∈ dom I → κ' ⊂ κ → κ' ∈ K') →
  Iinv -∗ lft_inv_alive κ (Vs κ) -∗ (∃ V, [†κ] V)
  ={userE ∪ ↑borN ∪ ↑inhN}=∗ Iinv ∗ lft_inv_dead κ (Vs κ).
Proof.
  iIntros (HVs Iinv HK HK') "(HI & Hdead & Halive) Hlalive Hκ".
  rewrite lft_inv_alive_unfold;
    iDestruct "Hlalive" as (P Q) "(Hbor & Hvs & Hinh)".
  rewrite /lft_bor_alive; iDestruct "Hbor" as (B) "[[Hbox Hbor] HB]".
  iAssert ⌜∀ i s, B !! i = Some s → ∃ V, s = Bor_in V ∧ V ⊑ (Vs κ)⌝%I with "[#]" as %HB.
  { iIntros (i s HBI). rewrite (big_sepM_lookup _ B) //.
    destruct s as [V|q|κ']; rewrite /bor_cnt; [iExists _; by iSplit| |].
    { iDestruct "HB" as "[_ HB]". iDestruct "Hκ" as (Vκ) "Hκ".
      set (V' := Vtok ⊔ Vκ).
      iDestruct (lft_tok_dead $! V' with "HB Hκ") as "[]". }
    iDestruct "HB" as "(% & % & Hcnt)". iDestruct (own_cnt_auth with "HI Hcnt") as %?.
    iDestruct (@big_sepS_elem_of with "Hdead") as "Hdead"; [by apply HK|].
    rewrite /lft_inv_dead; iDestruct "Hdead" as (R Vinh) "(_ & _ & Hcnt' & _)".
    iDestruct (own_cnt_valid_2 with "Hcnt' Hcnt")
      as %[?%nat_included _]%auth_both_valid_discrete; lia. }
  iMod (box_empty with "Hbox") as "[HP Hbox]"=>//; first set_solver+.
  { intros i s. by rewrite lookup_fmap fmap_Some=>-[? [/HB [?[-> ?]] ->]] /=. }
  rewrite lft_vs_unfold; iDestruct "Hvs" as (Vκ n) "(% & Hcnt & Hvs)".
  iDestruct (big_sepS_filter_acc (.⊂ κ) _ _ (dom I) with "Halive")
    as "[Halive Halive']"; [by eauto|].
  iApply fupd_trans. iApply (fupd_mask_mono (userE ∪ ↑borN)); [set_solver+|].
  iMod ("Hvs" $! I Vs with "[//] [//] [HI Halive] HP Hκ") as "(Hinv & HQ & Hcnt')".
  { rewrite lft_vs_inv_unfold. iFrame. }
  rewrite lft_vs_inv_unfold; iDestruct "Hinv" as "(HI&Halive)".
  iSpecialize ("Halive'" with "Halive").
  iMod (own_cnt_update_2 with "Hcnt Hcnt'") as "?".
  { apply auth_update_dealloc, (nat_local_update _ _ 0 0); lia. }
  rewrite /Iinv. iFrame "Hdead Halive' HI".
  iModIntro. iMod (lft_inh_kill with "[$Hinh $HQ]"); first set_solver+.
  iModIntro. rewrite /lft_inv_dead. iExists Q, (Vs κ). iFrame. iSplitR; [done|].
  rewrite /lft_bor_dead.
  iExists ((λ st, match st with Bor_in V => V | _ => inhabitant end) <$> B), P.
  rewrite -!map_fmap_compose.
  unfold to_borUR. rewrite (map_fmap_ext _ Excl); [by iFrame|naive_solver].
Qed.

Lemma lfts_kill A (I : gmap lft lft_names) (K K' : gset lft) (Vs : lft → Lat) :
  let Iinv K' := (own_ilft_auth I ∗ [∗ set] κ' ∈ K', lft_inv_alive κ' (Vs κ'))%I in
  (∀ κ κ', κ ∈ K → κ' ∈ dom I → κ' ⊆ κ → Vs κ ⊑ Vs κ') →
  K ## K' → dom I = K' ∪ K  →
  (∀ κ κ', κ ∈ K → κ' ∈ dom I → κ ⊆ κ' → κ' ∈ K) →
  Iinv K' -∗ ([∗ set] κ ∈ K,
     (∃ V, [†κ] V) ∗ lft_inv_alive κ (Vs κ) ∗ ⌜@lft_alive_in Lat A κ⌝ ∨
     lft_inv_dead κ (Vs κ) ∗ ⌜lft_dead_in A κ⌝)
    ={userE ∪ ↑borN ∪ ↑inhN}=∗ Iinv K' ∗ [∗ set] κ ∈ K, lft_inv_dead κ (Vs κ).
Proof.
  intros Iinv HVs. revert K'.
  induction (set_wf K) as [K _ IH]=> K' HKK' HKK'I HK.
  iIntros "[HI Halive] HK".
  pose (Kalive := filter (lft_alive_in A) K).
  destruct (decide (Kalive = ∅)) as [HKalive|].
  { iModIntro. rewrite /Iinv. iFrame.
    iApply (@big_sepS_impl with "[$HK]"); iModIntro.
    iIntros (κ Hκ) "[(_ & _ & %)|[$ _]]". set_solver. }
  destruct (minimal_exists_L (⊂) Kalive)
    as (κ & [Hκalive HκK]%elem_of_filter & Hκmin); first done.
  iDestruct (@big_sepS_delete with "HK") as "[Hκinv HK]"; first done.
  iDestruct "Hκinv" as "[(Hκ & Hκalive & _)|[_ %]]"; last first.
  { by destruct (lft_alive_and_dead_in A κ). }
  assert (κ ∉ K') as HκK' by set_solver +HκK HKK'.
  specialize (IH (K ∖ {[ κ ]})). feed specialize IH; [set_solver +HκK|set_solver +HVs|].
  iMod (IH ({[ κ ]} ∪ K') with "[HI Halive Hκalive] HK") as "[[HI Halive] Hdead]".
  { set_solver +HKK'. }
  { rewrite [_ ∪ K'](comm (R:=eq)) -(assoc (R:=eq) (∪)) -union_difference_L //
            -elem_of_subseteq_singleton //. }
  { intros κ' κ''. rewrite !elem_of_difference !elem_of_singleton=> -[? Hneq] ??.
    split; [by eauto|]; intros ->.
    eapply (minimal_strict_1 _ _ κ' Hκmin), strict_spec_alt; eauto.
    apply elem_of_filter; eauto using lft_alive_in_subseteq. }
  { rewrite /Iinv big_sepS_insert //. iFrame. }
  iDestruct (@big_sepS_insert with "Halive") as "[Hκalive Halive]"; first done.
  iMod (lft_kill with "[$HI $Halive $Hdead] Hκalive Hκ")
    as "[(HI&Halive&Hdead) Hκdead]".
  { auto. }
  { intros κ' ? [??]%strict_spec_alt.
    rewrite elem_of_difference elem_of_singleton; eauto. }
  { intros κ'. rewrite HKK'I elem_of_union=>-[//|Hκ'K Hκ'κ].
    destruct (minimal_strict_1 _ _ κ' Hκmin)=>//. rewrite elem_of_filter.
    split; [|done]. eapply lft_alive_in_subseteq, gmultiset_subset_subseteq; eauto. }
  iModIntro. rewrite /Iinv (big_sepS_delete _ K) //. iFrame.
Qed.

Lemma exists_Vs A (K : gset lft) :
  ([∗ set] κ ∈ K, lft_inv A κ) -∗
  ∃ Vs,
    ⌜(∀ κ, κ ∈ K → lft_has_view A κ (Vs κ))⌝ ∗
    ⌜(∀ κ κ', κ ∈ K → κ' ∈ K → κ' ⊆ κ → Vs κ ⊑ Vs κ')⌝ ∗
    ([∗ set] κ ∈ K, lft_inv_alive κ (Vs κ) ∗ ⌜lft_alive_in A κ⌝ ∨
                    lft_inv_dead κ (Vs κ) ∗ ⌜lft_dead_in A κ⌝).
Proof.
  induction (set_wf K) as [K _ IH]. iIntros "HK".
  destruct (decide (K = ∅)) as [->|].
  { iExists (λ _, inhabitant). repeat (iSplit; [by auto|]). by rewrite !big_sepS_empty. }
  destruct (minimal_exists_L (⊂) K) as (κ & HκK & Hκmin); first done.
  rewrite big_sepS_delete //. iDestruct "HK" as "[Hinv HK]".
  iDestruct (IH with "HK") as (Vs) "(Hhv0 & Hmono & HK)". { set_solver +HκK. }
  iDestruct "Hmono" as %Hmono.
  rewrite /lft_inv. iDestruct "Hinv" as (Vκ) "[Hhv Hinv]".
  iDestruct "Hhv" as %Hhv. iDestruct "Hhv0" as %Hhv0.
  unshelve iExists (<[κ := Vκ ⊔ foldr (⊔) ε (Vs <$> elements (filter (κ ⊂.) K))]>Vs);
    try apply _; [shelve..|].
  repeat iSplit.
  - iPureIntro. intros κ' Hκ'K Λ HΛ. destruct (decide (κ = κ')) as [<-|].
    + rewrite functions.fn_lookup_insert. specialize (Hhv Λ HΛ).
      destruct (A !! Λ) as [[b VΛ]|] eqn:EQVΛ=>//=. apply lat_join_lub; [done|].
      assert (HFA : Forall (λ κ', Vs κ' ⊑ VΛ) (elements (filter (κ ⊂.) K))).
      { apply set_Forall_elements. intros κ'. rewrite elem_of_filter=>-[Hκκ' Hκ'K'].
        specialize (λ H, Hhv0 κ' H Λ). rewrite EQVΛ in Hhv0. apply Hhv0.
        - rewrite elem_of_difference elem_of_singleton. split; [done|intros ->].
            by eapply strict_ne.
        - eapply gmultiset_elem_of_subseteq; [done|]. apply Hκκ'. }
      clear -HFA. clear IH. induction elements as [|?? IH]=>//=;
         [|inversion_clear HFA]; solve_lat.
    + rewrite functions.fn_lookup_insert_ne //. apply Hhv0, HΛ.
      rewrite elem_of_difference elem_of_singleton. auto.
  - iPureIntro. intros κ1 κ2 Hκ1 Hκ2 Hκ1κ2.
    destruct (decide (κ1 = κ)) as [->|], (decide (κ2 = κ)) as [->|];
      rewrite ?functions.fn_lookup_insert ?functions.fn_lookup_insert_ne //.
    + destruct (Hκmin κ2) as [_ []]=>//. apply strict_spec_alt. auto.
    + rewrite <-lat_join_sqsubseteq_r.
      assert (Hκ1K : κ1 ∈ elements (filter (κ ⊂.) K)).
      { apply elem_of_elements. rewrite elem_of_filter.
        split; [|done]. apply strict_spec_alt. auto. }
      clear -Hκ1K. induction Hκ1K; solve_lat.
    + apply Hmono=>//; rewrite elem_of_difference elem_of_singleton; auto.
  - rewrite (big_sepS_delete _ K) //. iSplitL "Hinv".
    + rewrite functions.fn_lookup_insert. by rewrite <-lat_join_sqsubseteq_l.
    + iApply (@big_sepS_impl with "[$HK]"). iModIntro. iIntros (κ').
      rewrite elem_of_difference elem_of_singleton. iIntros ([??]) "H".
        rewrite functions.fn_lookup_insert_ne //.
Qed.

Lemma lft_create_strong P E :
  pred_infinite P → ↑lftN ⊆ E →
  ⎡lft_ctx⎤ ={E}=∗
  ∃ p : positive, let κ := positive_to_lft p in ⌜P p⌝ ∗
       (1).[κ] ∗ □ ((1).[κ] ={↑lftN ∪ userE}[userE]▷=∗ [†κ]).
Proof.
  assert (userE_lftN_disj:=userE_lftN_disj). iIntros (HP ?) "#LFT".
  iInv mgmtN as (A I) "(>HA & >HI & Hinv)" "Hclose".
  rewrite ->(pred_infinite_set (C:=gset _)) in HP.
  destruct (HP (dom A)) as [Λ [HPx HΛ%not_elem_of_dom]].
  iMod (own_update with "HA") as "[HA HΛ]".
  { apply auth_update_alloc, (alloc_singleton_local_update _ Λ
           (Cinl (1%Qp, to_latT bot)))=>//. by rewrite lookup_fmap HΛ. }
  iMod ("Hclose" with "[HA HI Hinv]") as "_".
  { iNext. rewrite /lfts_inv /own_alft_auth.
    iExists (<[Λ:=(true, bot)]>A), I. rewrite /to_alftUR fmap_insert; iFrame.
    rewrite big_sepS_impl. iApply "Hinv".
    rewrite /lft_inv. iIntros (κ ?) "!> !> H". iDestruct "H" as (Vκ) "[Hhv H]".
    iDestruct "Hhv" as %Hhv. iExists Vκ. iSplit.
    - iPureIntro. intros Λ' HΛ'. specialize (Hhv Λ' HΛ').
      rewrite lookup_insert_ne //. intros <-. by rewrite HΛ in Hhv.
    - iDestruct "H" as "[[Hκ %]|[Hκ %]]".
      + iLeft. iFrame "Hκ". iPureIntro. by apply lft_alive_in_insert.
      + iRight. iFrame "Hκ". iPureIntro. by apply lft_dead_in_insert. }
  iModIntro; iExists Λ. iSplit; first done. iSplitL.
  { rewrite /lft_tok big_sepMS_singleton. iExists _. iFrame.
    iDestruct (monPred_in_intro True%I with "[//]") as (V) "[H _]".
    by rewrite -(lat_bottom_sqsubseteq V). }
  clear I A HΛ. iIntros "!> HΛ".
  iApply (step_fupd_mask_mono (↑lftN ∪ userE) _ ((↑lftN ∪ userE)∖↑mgmtN)); [solve_ndisj..|].
  iInv mgmtN as (A I) "(>HA & >HI & Hinv)" "Hclose".
  iMod (ilft_create _ _ {[+ Λ +]} with "HA HI Hinv") as "H". clear A I.
  iDestruct "H" as (A I) "(% & HA & HI & Hinv)".
  rewrite /lft_tok /= big_sepMS_singleton. iDestruct "HΛ" as (V'0) "[#HV'0 HΛ]".
  iDestruct (own_valid_2 with "HA HΛ")
    as %[[s [Hs Hincl%Some_included]]%singleton_included_l Hv]%auth_both_valid_discrete.
  destruct Hincl; last first.
  { destruct (exclusive_included (Cinl (1%Qp, to_latT V'0)) s)=>//.
    move: (Hv Λ). by rewrite Hs. }
  setoid_subst. revert Hs.
  case Eqs : (to_alftUR A !! Λ)=>[[[]| |]|];
    (try by inversion 1); intros EQ%(inj Some); try by inversion EQ.
  apply (reflexive_eq (R := (≡))) in Eqs.
  apply (inj Cinl), (inj2 pair) in EQ. destruct EQ. setoid_subst.
  assert (EQAΛ : A !! Λ ≡ Some (true, V'0)).
  { rewrite lookup_fmap in Eqs. revert Eqs.
    case: (A!!Λ)=>[[[]?]|]; (try by inversion 1); intros EQ%(inj Some).
    - by apply (inj Cinl), (inj2 pair) in EQ; do 2 f_equiv; naive_solver.
    - by inversion EQ. }
  iDestruct (exists_Vs with "Hinv") as (Vs) "(>% & >% & Hinv)".
  iMod (own_update_2 with "HA HΛ") as "[HA HΛ]".
  { apply Some_equiv_eq in Eqs as (? & ? & ?).
    by eapply auth_update, singleton_local_update,
       (exclusive_local_update _ (Cinr (to_agree (to_latT (Vs {[Λ]}))))). }
  iDestruct "HΛ" as "#HΛ". iModIntro; iNext.
  pose (K' := filter (λ κ, lft_alive_in A κ ∧ Λ ∉ κ) (dom I)).
  destruct (proj1 (subseteq_disjoint_union_L K' (dom I))) as (K&HI&HK).
  { set_solver+. }
  assert (elem_of_K : ∀ κ, κ ∈ K ↔ κ ∈ dom I ∧ (lft_dead_in A κ ∨ Λ ∈ κ)).
  { rewrite HI=>κ. rewrite elem_of_union elem_of_filter. split.
    - intros HκK. split; [by auto|]. destruct (decide (Λ ∈ κ)) as [|HΛκ]; [by auto|left].
      destruct (lft_alive_or_dead_in A κ) as [(Λ'&HΛ'&EQΛ')|[Hal|?]]=>//.
      + exfalso. assert (κ ∈ dom I). { by set_solver+HI HκK. }
        assert (lft_has_view A κ (Vs κ)) as Hhv' by auto.
        specialize (Hhv' Λ' HΛ'). by rewrite EQΛ' in Hhv'.
      + edestruct HK=>//. set_solver + Hal HΛκ HκK HI.
    - move=>-[[[[??]?]|?][?|?]]//. by destruct (lft_alive_and_dead_in A κ). }
  rewrite HI !big_sepS_union //. iDestruct "Hinv" as "[HinvD HinvK]".
  iApply fupd_trans. iApply (fupd_mask_mono (userE ∪ ↑borN ∪ ↑inhN)); first solve_ndisj.
  iMod (lfts_kill A I K K' Vs with "[$HI HinvD] [HinvK]") as "[[HI HinvD] HinvK]"=>//.
  { intros κ κ' Hκ Hκ' Hκκ'. assert (κ ∈ dom I) by set_solver +HI Hκ. auto. }
  { move=> κ κ'. rewrite !elem_of_K=>-[?[Hd|?]] ??; split=>//.
    - left. destruct Hd as (Λ'&?&?). exists Λ'. split; [|done].
      by eapply gmultiset_elem_of_subseteq.
    - right. by eapply gmultiset_elem_of_subseteq. }
  { rewrite big_sepS_impl. iApply "HinvD".
    iIntros "!> !>" (κ [[? _]_]%elem_of_filter) "[[$ _]|[_ %]]".
    by edestruct @lft_alive_and_dead_in. }
  { rewrite big_sepS_impl. iApply "HinvK".
    iIntros "!> !>" (κ [_ Hκ]%elem_of_K) "[[H %]|H]"; [iLeft|iRight]; iFrame "H %".
    rewrite /lft_dead /=. iExists _, _, _. iIntros "{$HΛ} !%". split; [|done].
    case Hκ=>// Hd. by destruct (lft_alive_and_dead_in A κ). }
  iModIntro. iMod ("Hclose" with "[-]") as "_"; last first.
  { iModIntro. rewrite /lft_dead. iExists Λ, _. iIntros "{$HΛ}".
    rewrite gmultiset_elem_of_singleton. iSplit; [done|].
    assert (Hhv : lft_has_view A {[+ Λ +]} (Vs {[+ Λ +]})) by auto.
    move: (Hhv Λ). rewrite gmultiset_elem_of_singleton. revert EQAΛ.
    case: (A!!Λ)=>[[??]|]; [|intros EQ; by inversion EQ]. simpl.
    intros [_ ->]%(inj Some)%(inj2 pair)=>/(_ eq_refl) <- //. }
  set (A' := <[Λ:=(false, Vs {[+ Λ +]})]>A).
  assert (Hhv : ∀ κ, κ ∈ dom I → lft_has_view A' κ (Vs κ)).
  { intros κ Hκ Λ' HΛ'. assert (Hhv : lft_has_view A κ (Vs κ)) by auto.
    specialize (Hhv Λ' HΛ').
    destruct (decide (Λ =  Λ')) as [<-|]; rewrite ?lookup_insert ?lookup_insert_ne //=.
    assert ({[Λ]} ⊆ κ) by by apply gmultiset_singleton_subseteq_l. auto. }
  iModIntro. iModIntro. iExists A', I.
  rewrite /own_alft_auth /to_alftUR fmap_insert. iFrame "HA HI".
  rewrite HI !big_sepS_union //. iSplitL "HinvD".
  - rewrite big_sepS_impl. iApply "HinvD".
    iIntros "!>" (κ [[??]?]%elem_of_filter) "Halive".
    rewrite /lft_inv. iExists (Vs κ). iSplitR; [by auto|].
    iLeft. iFrame "Halive". iPureIntro. by apply lft_alive_in_insert_false.
  - rewrite big_sepS_impl /lft_inv. iApply "HinvK".
    iIntros "!>" (κ [? HκK]%elem_of_K) "Hdead". iExists (Vs κ). iSplitR; [by auto|].
    iRight. iFrame. iPureIntro. destruct HκK.
    + by apply lft_dead_in_insert_false. + by apply lft_dead_in_insert_false'.
Qed.
End creation.
